package com.dzdy.dim.protocol;

import com.dzdy.dim.protocol.command.Command;
import com.dzdy.dim.serialize.Serializer;
import com.dzdy.dim.serialize.impl.JSONSerializer;
import io.netty.buffer.ByteBuf;
import io.netty.buffer.ByteBufAllocator;

import java.util.HashMap;
import java.util.Map;

/**
 * 通信包编码
 * 魔数(4)+版本号(1)+序列化算法(1)+指令(1)+数据长度(4)+数据(N)
 *
 * @author : wangzhiyong
 * @date : 2019/5/25 15:39
 * description :
 */
public class PacketCode {

    /**
     * 单例
     */
    public static final PacketCode INSTANCE = new PacketCode();
    /**
     * 魔数 标识自定义通讯包 不是http数据包 节省资源
     */
    public static final int MAGIC_NUMBER = 0x7776447;
    /**
     * 序列化方式集合
     */
    private static final Map<Byte, Serializer> SERIALIZER_MAP;

    static {
        SERIALIZER_MAP = new HashMap<>();
        Serializer serializer = new JSONSerializer();
        SERIALIZER_MAP.put(serializer.getSerializerAlgorithm(), serializer);
    }

    /**
     * 编码
     *
     * @param packet
     * @return
     */
    public ByteBuf encode(Packet packet) {
        ByteBuf byteBuf = ByteBufAllocator.DEFAULT.buffer();
        byte[] bytes = Serializer.DEFAULT.serialize(packet);
        encodeData(byteBuf, packet, bytes);
        return byteBuf;
    }

    /**
     * 编码
     *
     * @param byteBuf
     * @param packet
     */
    public void encode(ByteBuf byteBuf, Packet packet) {
        byte[] bytes = Serializer.DEFAULT.serialize(packet);
        encodeData(byteBuf, packet, bytes);
    }

    private void encodeData(ByteBuf byteBuf, Packet packet, byte[] bytes) {
        byteBuf.writeInt(MAGIC_NUMBER);
        byteBuf.writeByte(packet.getVersion());
        byteBuf.writeByte(Serializer.DEFAULT.getSerializerAlgorithm());
        byteBuf.writeByte(packet.getCommand());
        byteBuf.writeInt(bytes.length);
        byteBuf.writeBytes(bytes);
    }

    /**
     * 解码
     *
     * @param byteBuf
     * @return
     */
    public Packet decode(ByteBuf byteBuf) {
        try {
            //跳过 magic number
            byteBuf.skipBytes(4);
            byteBuf.skipBytes(1);

            // 序列化算法标识
            byte serializeAlgorithm = byteBuf.readByte();
            // 指令
            byte command = byteBuf.readByte();
            // 数据包长度
            int length = byteBuf.readInt();

            //将数据写入字节数组
            byte[] bytes = new byte[length];
            byteBuf.readBytes(bytes);

            Command comm = Command.getCommand(command);
            Serializer serializer = getSerializer(serializeAlgorithm);
            if (null != serializer && null != comm) {
                Packet packet = serializer.deserialize(bytes, comm.getClazz());
                if (packet instanceof SockData) {
                    SockData sockData = (SockData) packet;
                    sockData.setCommand(comm);
                    return sockData;
                }
                return packet;
            }
        } catch (Exception e) {
            System.out.println("数据异常，解析失败");
        }
        return null;
    }

    private Serializer getSerializer(byte serializeAlgorithm) {
        return SERIALIZER_MAP.get(serializeAlgorithm);
    }
}
